"use strict";
/*
    Migrate.js - OneTable Migrations
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Migrate = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const semver_1 = __importDefault(require("semver"));
const dynamodb_onetable_1 = require("dynamodb-onetable");
// DEV
// import { Model, Table } from '../../onetable/dist/mjs/index.js'
//  Cache of key and delimiter information per table
const TableCache = {};
class Migrate {
    /*
        Construct the Migrate instance.
        The `config` parameter is either instance of Table() or params for Table.
        The `params.migrations` set to the in-memory migrations or `params.dir` to a migrations directory.
     */
    constructor(config = {}, params = {}) {
        this.params = params;
        this.migrations = params.migrations;
        this.dir = path_1.default.resolve(params.dir || '.');
        this.db = (typeof config.setClient != 'function') ? new dynamodb_onetable_1.Table(config) : config;
    }
    getTableInfo() {
        return __awaiter(this, void 0, void 0, function* () {
            let db = this.db;
            let info = TableCache[db.name];
            if (info && info.updated > (Date.now() - 60 * 1000)) {
                return info;
            }
            let data = yield db.describeTable();
            info = {};
            for (let key of data.Table.KeySchema) {
                let type = key.KeyType.toLowerCase() == 'hash' ? 'hash' : 'sort';
                info[type] = key.AttributeName;
            }
            if (!info.sort) {
                throw new Error('Cannot use Migrate library on a table without a sort/range key');
            }
            info.updated = Date.now();
            TableCache[db.name] = info;
            return info;
        });
    }
    getModel() {
        return __awaiter(this, void 0, void 0, function* () {
            let db = this.db;
            let info = yield this.getTableInfo();
            //  Read the schema to update the Table() params
            let schema = yield db.readSchema();
            if (schema) {
                db.setSchema(schema);
            }
            return this.db.getModel('_Migration');
        });
    }
    /*
        Invoke a migration. Method is up or down.
        This will set the schema if the migration defines one and will persist the schema to the table after the migration.
    */
    invoke(migration, method) {
        return __awaiter(this, void 0, void 0, function* () {
            let db = this.db;
            if (migration.schema) {
                db.setSchema(migration.schema);
            }
            yield migration[method](db, this);
            if (method == 'down') {
                let current = yield this.loadMigration(yield this.getCurrentVersion());
                if (current.schema || migration.schema) {
                    yield db.saveSchema(current.schema || migration.schema);
                }
            }
            else if (migration.schema) {
                db.saveSchema(migration.schema);
            }
        });
    }
    /* public */
    apply(direction, version) {
        return __awaiter(this, void 0, void 0, function* () {
            let db = this.db;
            let Migration = yield this.getModel();
            let migration;
            if (direction == 0) {
                //  Reset to zero
                yield Migration.remove({}, { many: true });
                //  Create prior migration items
                let versions = yield this.getVersions();
                version = versions.pop();
                for (let v of versions) {
                    migration = yield this.loadMigration(v);
                    yield Migration.create({
                        version: v,
                        date: new Date(),
                        path: migration.path,
                        description: migration.description,
                    });
                }
            }
            migration = yield this.loadMigration(version);
            if (direction < 0) {
                yield this.invoke(migration, 'down');
                yield Migration.remove({ version: migration.version });
            }
            else {
                //  Up, repeat or reset
                yield this.invoke(migration, 'up');
                let params = {
                    version,
                    date: new Date(),
                    path: migration.path,
                    description: migration.description,
                };
                if (direction <= 1) {
                    yield Migration.create(params);
                }
                else {
                    yield Migration.update(params);
                }
            }
            return migration;
        });
    }
    /* public */
    findPastMigrations() {
        return __awaiter(this, void 0, void 0, function* () {
            let Migration = yield this.getModel();
            let pastMigrations = yield Migration.find({});
            this.sortMigrations(pastMigrations);
            return pastMigrations;
        });
    }
    /* public */
    getCurrentVersion() {
        return __awaiter(this, void 0, void 0, function* () {
            let pastMigrations = yield this.findPastMigrations();
            if (pastMigrations.length == 0) {
                return '0.0.0';
            }
            return pastMigrations[pastMigrations.length - 1].version;
        });
    }
    /*
        Return outstanding versions in semver sorted order up to the specified limit
     */
    /* public */
    getOutstandingVersions(limit = Number.MAX_SAFE_INTEGER) {
        return __awaiter(this, void 0, void 0, function* () {
            let current = yield this.getCurrentVersion();
            let pastMigrations = yield this.findPastMigrations();
            let versions = yield this.getVersions();
            versions = versions.filter(version => {
                return semver_1.default.compare(version, current) > 0;
            }).sort(semver_1.default.compare);
            versions = versions.filter(v => pastMigrations.find(m => m.version == v) == null);
            return versions.slice(0, limit);
        });
    }
    /*
        Return outstanding migrations in semver sorted order up to the specified limit
     */
    /* public */
    getOutstandingMigrations(limit = Number.MAX_SAFE_INTEGER) {
        return __awaiter(this, void 0, void 0, function* () {
            let versions = yield this.getOutstandingVersions(limit);
            let migrations = [];
            for (let v of versions) {
                migrations.push(yield this.loadMigration(v));
            }
            return migrations;
        });
    }
    /* private */
    getVersions() {
        return __awaiter(this, void 0, void 0, function* () {
            let versions;
            if (this.migrations) {
                versions = this.migrations.map(m => m.version);
            }
            else {
                versions = fs_1.default.readdirSync(this.dir).map(file => file.replace(/\.[^/.]+$/, ''));
            }
            return versions.filter(version => { return semver_1.default.valid(version); }).sort(semver_1.default.compare);
        });
    }
    /* private */
    loadMigration(version) {
        return __awaiter(this, void 0, void 0, function* () {
            let path, task;
            if (this.migrations) {
                task = this.migrations.find(m => m.version == version);
            }
            else {
                path = `${this.dir}/${version}.js`;
                task = (yield Promise.resolve().then(() => __importStar(require(path)))).default;
            }
            if (!task) {
                throw new Error(`Cannot find migration for version ${version}`);
            }
            if (!task.schema) {
                throw new Error(`Migration ${version} is missing a schema`);
            }
            if (!task.version) {
                throw new Error(`Migration ${version} is missing a version property`);
            }
            return {
                version,
                description: task.description,
                path: path || 'memory',
                schema: task.schema,
                up: task.up,
                down: task.down,
            };
        });
    }
    /* private */
    sortMigrations(array) {
        array.sort((a, b) => {
            let cmp = semver_1.default.compare(a.version, b.version);
            if (cmp < 0) {
                return cmp;
            }
            else if (cmp > 0) {
                return cmp;
            }
            else if (a.order < b.order) {
                return -1;
            }
            else if (a.order > b.order) {
                return 1;
            }
            else {
                return 0;
            }
        });
    }
}
exports.Migrate = Migrate;
